	.TITLE	IOB6120 I/O Board extension ROM

;   Copyright (C) 2003 by J. Kearney, Bolton, Massachusetts
;   Portions adapted from and Copyrighted (C) 2001-2003
;   by R. Armstrong, Milpitas, California.
;
;   This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
;   This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
; for more details.
;
;   You should have received a copy of the GNU General Public License along
; with this program; if not, write to the Free Software Foundation, Inc.,
; 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;

	.HM6120
	.STACK  PAC1, POP1, PPC1, RTN1
	.NOWARN	F

	.TITLE	Edit History

;
;  100		-- Beginning of recorded history (end of alpha development)
;
;  101		-- Added FPGA VHDL version display
;
;  102		-- Added disassembler, X and XP commands, hooked HELP and
;                  TYPEIR.
;
VERSION=102

	.Title Constants and other definitions

;
; Configuration constants
;

BTSMINVERSION=226	; minimum version of BTS6120 that's acceptable
CFPARITION=7700		; CF partitions start here

;
; SBC6120 Memory map control for panel RAM access
;
		;   DIRECT	INDIRECT
		;   --------	--------
MM0=6400	;   EPROM	RAM	(automatically selected by a RESET)
MM1=6401	;   RAM		EPROM	(used during system initialization)
MM2=6402	;   RAM		RAM	(used almost all the time)
MM3=6403	;   RAM		DISK	(used to access RAM disk only)

;
; SBC6120 RAMdisk address register access
;

LDAR=6410

;
; SBC6120 IOTs to control serial port address
;

PRISLU=6412
SECSLU=6413
KCC=6032	; Clear receive flag and AC
TSF=6041	; Skip if the console transmit flag is set
TLS=6046	; Load AC into transmit buffer and clear the flag

;
; Mapping of EMAs to subsystems
;

CDF_ROM	   = CDF 1
CDF_CF	   = CDF 3
CDF_FPGA   = CDF 5
CDF_STATUS = CDF 7

CDF_BTS0   = CDF 0
CDF_BTS1   = CDF 1
CDF_HERE   = CDF 2
CIF_HERE   = CIF 2
CXF_HERE   = CXF 2

;
; CompactFlash registers
;

CF_DATA	= 0			; ADDRESS OF DATA REGISTER
CF_ERR	= 1			; ADDRESS OF ERROR REGISTER
CF_FEAT	= 1			; ADDRESS OF FEATURES REGISTER
CF_CNT	= 2			; ADDRESS OF SECTOR COUNT REGISTER
CF_LB0	= 3			; ADDRESS OF LB0-7 / SECTOR NUMBER REGISTER
CF_LB1	= 4			; ADDRESS OF LB8-15 / LOW CYLINDER REGISTER
CF_LB2	= 5			; ADDRESS OF LB16-23 / HIGH CYLINDER REGISTER
CF_LB3	= 6			; ADDRESS OF LB24-27 / HEAD/DRIVE REGISTER
CF_STAT	= 7			; ADDRESS OF STATUS REGISTER
CF_CMD	= 7			; ADDRESS OF COMMAND REGISTER

; IDE status register (REGSTS) bits...
STSBSY=0200	; busy
STSRDY=0100	; device ready
STSDF= 0040	; device fault
STSDSC=0020	; device seek complete
STSDRQ=0010	; data request
STSCOR=0004	; corrected data flag
STSERR=0001	; error detected

; IDE command codes (or at least the ones we use!)...
CMDIDD=354	; identify device
CMDRDS=040	; read sectors with retry
CMDWRS=060	; write sectors with retry

;
; Bit definitions on status port
;

STAT_INIT=0002
STAT_DONE=0001
STAT_JUMPER=0010
STAT_CFDETECT=0020
STAT_CFRDY=0040

;
; FPGA IOTs
;

FVERSION=6422
FUNLOCK=6423
FSETCON=6424
FREPROG=6425
TESTSIG=6426
TESTBUS=6427

KEY=5253

;
; Vector table locations.  Encoded as (Offset * 2) + Field_Number
;

; Field 0 vectors
VRET0F2	=  0.
VUPC	=  2.			; V User PC
VUAC	=  4.			; V User AC
VUFLAGS	=  6.			; V User Flags
VSYSIN9	=  8.
VCONOUT	= 10.
VCONIN	= 12.
VOUTCHR	= 14.
VINCHRS	= 16.
VTOCT4	= 18.
VCRLF	= 20.
VSPACMP	= 22.
VSPACM0	= 24.
VBACKUP	= 26.
VEOLTST	= 28.
VEOLNXT	= 30.
VGET	= 32.
VOCTNW	= 34.
VRANGE	= 36.
VRESTA	= 38.
VCOMERR	= 40.
VTDECNW	= 42.
VNODISK	= 44.
VTYPEIR	= 46.			; V  print the current IR
VUIR	= 48.			; V  user IR
VADDR	= 50.			; V  start address returned from RANGE
VADDRFLD= 52.			; V  start field returned from RANGE
VHIGH	= 54.			; V  end address returned from RANGE
VHIGHFLD= 56.			; V  end field returned from RANGE
; Field 1 vectors
VRET1F2	=  1.
VFUNTBL	= 3.			; V  table of ROM functions
VCMDTBL	= 5.			; V  table of monitor commands
VDISKRD	= 7.			; H  read IDE
VDISKWR	= 9.			; H  write IDE
VPARMAP	= 11.			; V  partition map
VDKPART	= 13.			; V  partition number for DISKRD/WR
VDKRBN	= 15.			; V  block number for DISKRD/WR
VBUFPTR	= 17.			; V  offset for DISKRD/WR
VBUFCDF	= 19.			; V  field for DISKRD/WR
VBUFPNL	= 21.			; V  memory space for DISKRD/WR
VDKSIZE	= 23.			; V  size of IDE attached drive

;
; OPR microinstructions that load the AC with various special constants...
;
NL0000=CLA			; all models
NL0001=CLA IAC         		; all models
NL0002=CLA CLL CML     RTL	; all models
NL2000=CLA CLL CML     RTR	; all models
NL3777=CLA CLL     CMA RAR	; all models
NL4000=CLA CLL CML     RAR	; all models
NL5777=CLA CLL     CMA RTR	; all models
NL7775=CLA CLL     CMA RTL	; all models
NLM3=NL7775			; all models
NL7776=CLA CLL     CMA RAL	; all models
NLM2=NL7776			; all models
NL7777=CLA         CMA		; all models
NLM1=NL7777			; all models
NL0003=CLA STL IAC RAL		; PDP-8/I and later
NL0004=CLA CLL IAC RTL		; PDP-8/I and later
NL0006=CLA STL IAC RTL		; PDP-8/I and later
NL6000=CLA STL IAC RTR		; PDP-8/I and later
NL0100=CLA IAC BSW		; PDP-8/E and later
NL0010=CLA IAC R3L		; HM6120 only

	.Title Variables
	.FIELD 2

	.ORG	0000

BTSVER:	.BLOCK	1
F0VECTOR: .BLOCK 1
F1VECTOR: .BLOCK 1
RAMPTR:	.BLOCK 1
PUSHAC:	.BLOCK	1
DESTA:	.BLOCK	1
POP0T2:	.BLOCK	1
POP1T2:	.BLOCK	1

	.ORG	0010		; autoindex variables

ARGPTR:	.BLOCK	1
F0PATCH: .BLOCK	1
F1PATCH: .BLOCK	1
AX0:	.BLOCK	1
AX1:	.BLOCK	1
AX2:	.BLOCK	1
ASCP:	.BLOCK	1		; pointer for TASCIZ

	.ORG	0020

; pointers to BTS6120 stuff we want to have available
; these will be set up by INITVECT ...
PRESTA:	.BLOCK	1		; monitor restart vector (field 0)
PDKSIZE: .BLOCK	1
PPARMAP: .BLOCK	1
PDKPART: .BLOCK	1
PDKRBN:	.BLOCK	1
PBUFPTR: .BLOCK	1
PBUFCDF: .BLOCK	1
PBUFPNL: .BLOCK	1
PDISKWR: .BLOCK	1
PDISKRD: .BLOCK	1
PNODISK: .BLOCK	1
PUIR:	.BLOCK	1		; User Instruction Register
PUPC:	.BLOCK	1		; User PC
PUFLAGS: .BLOCK	1		; User Flags
PTYPEIR: .BLOCK	1		; H  print the current IR
PADDR:	.BLOCK	1		; V  start address returned from RANGE
PADDRFLD: .BLOCK 1		; V  start field returned from RANGE
PHIGH:	.BLOCK	1		; V  end address returned from RANGE
PHIGHFLD: .BLOCK 1		; V  end field returned from RANGE
; .. to here

X0:	.BLOCK	1
X1:	.BLOCK	1
X2:	.BLOCK	1
X3:	.BLOCK	1

FPGAERR: .BLOCK	1
CFSIZE:	.BLOCK	1		; 0 if not present, else MB

BUFSIZ:	.BLOCK	1
CFPART:	.BLOCK	1
CFRBN:	.BLOCK	1

RWRET:	.BLOCK	1
CFTO:	.BLOCK	2

BTSHELP: .BLOCK	1

; dissembler vars
IR:	.BLOCK	1		; instruction
PC:	.BLOCK	1		; location
DFIF:	.BLOCK	1		; IF in b6-8, DF in b9-11
DPNL:	CPD			; [SPD] or [CPD]

	.TITLE Page Zero Utility Functions

	.ORG	0100

GETV:	0
	CLL
	RAR			; low bit into Link
	SZL
	 JMP	GETV1
GETV0:	TAD	F0VECTOR	; add base
	DCA	X0		; save vector address and clear AC
	CDF_BTS0
	TAD	@X0
	CDF_HERE
	JMP	@GETV
GETV1:	TAD	F1VECTOR	; add base
	DCA	X0		; save vector address and clear AC
	CDF_BTS1
	TAD	@X0
	CDF_HERE
	JMP	@GETV

;   This two routines will allow a routine in field 2 to simulate a .PUSHJ
; to a routines in field 0 or 1. The contents of the AC are preserved both
; ways across the call.
;
;CALL:
;	JMS	PJ0F2 or PJ1F2	; cross field call "PUSHJ to 0 From 2"
;	 <addr>			; address of a routine in field 1
;	<return here>		; with the AC preserved across the call
;
PJ0F2:	0			; call here with a JMS instruction
	DCA	PUSHAC		; save the caller's AC for a minute
	CDF_HERE
	TAD	@PJ0F2		; then get caller's argument
	DCA	DESTA		; that's the address of the routine to call
	TAD	PJ0F2		; now get caller's return address
	IAC			; and skip over the argument
	.PUSH			; put that on the stack
	CLA			; (PUSH doesn't clear the AC!)
	TAD	POP0T2		; the field 0 routine will return to
	.PUSH			; ...
	CLA			; ...
	TAD	PUSHAC		; restore the original AC contents
	CXF	0		; call with IF = DF = 0
	JMP	@DESTA		; and go to the code in field 0

PJ1F2:	0			; call here with a JMS instruction
	DCA	PUSHAC		; save the caller's AC for a minute
	CDF_HERE
	TAD	@PJ1F2		; then get caller's argument
	DCA	DESTA		; that's the address of the routine to call
	TAD	PJ0F2		; now get caller's return address
	IAC			; and skip over the argument
	.PUSH			; put that on the stack
	CLA			; (PUSH doesn't clear the AC!)
	TAD	POP1T2		; the field 1 routine will return to
	.PUSH			; ...
	CLA			; ...
	TAD	PUSHAC		; restore the original AC contents
	CXF	1		; call with IF = DF = 1
	JMP	@DESTA		; and go to the code in field 1

	.Title Initialization
;
; Extension ROM initialization entry point.  Called with AC containing the
; number of parameter words following the call.  This code is slightly
; tricky in that it is overlaid onto the variable area, and so care must
; be taken not to set any variables at locations that have not been executed yet

; also important: this entry point is called from page 1, and the parameters
; of course are there

	; this stuff overlays the first couple of variables
	.ORG	0

	.SIXBIT	/SBC6/		; 2 word signature, soon to be overwritten
IOBIN0: 0
	JMP	@[IOBIN1]

	.PAGE	1

CHKSUM:	.BLOCK	1

; first, build the return address and set up our parameter pointer

IOBIN1:	TAD	IOBIN0		; AC=num params, so now points past end of them
	.PUSH			; ... which is our return address

	NL7777
	TAD	IOBIN0		; get start of params - 1 for autoincr
	DCA	ARGPTR

; second, copy the parameters

	TAD	@ARGPTR		; fetch version number
	DCA	BTSVER		; stash it away for now

	TAD	@ARGPTR		; fetch and save F0VECTOR
	DCA	F0VECTOR
	TAD	@ARGPTR		; fetch and save F0PATCH-1
	DCA	F0PATCH

	TAD	@ARGPTR		; fetch and save F1VECTOR
	DCA	F1VECTOR
	TAD	@ARGPTR		; fetch and save F1PATCH-1
	DCA	F1PATCH

	CDF_BTS0
	TAD	@F0VECTOR	; get 1st vector to .POPJ hook
	DCA	POP0T2		; so that we can call Field 0 routines
	CDF_BTS1
	TAD	@F1VECTOR	; get 1st vector to .POPJ hook
	DCA	POP1T2		; so that we can call Field 1 routines

; now call the various initializations.  If anything fails, return an
; error code to BTS6120

	TAD	BTSVER		; get back the version
	TAD	[-BTSMINVERSION] ; check it against our minimum requirement
	SMA CLA
	 JMP	 IOBIN2		; okay, same or newer than that
	NL0003			; error code 0003 - bad BTS version
	JMP	XINIT

IOBIN2:
	.PUSHJ	@[INITVECT]	; link calls into BTS from this code

	CDF_BTS0		; access Field 0

	; n.b. autoincr AX0 was set up by INITVECT to point to SYSINI9-1
	TAD	[CXF_HERE]		; set up CXF 2 / JMP @[.+1] / INI9HOOK
	DCA	@AX0

	NL0002			; compute address of constant INI9HOOK
	TAD	AX0
	AND	[0177]		; make into a JMP @.+1
	TAD	[5600]
	DCA	@AX0

	TAD	[INI9HOOK]	; and store the address
	DCA	@AX0

	CDF_HERE

	.PUSHJ	@[INITCFC]	; initialize CF
	.PUSHJ	@[CFHOOKS]	; hook up CFC

	.PUSHJ	@[IFPGA]	; initialize FPGA
	SZA			; on error, return
	 JMP	 XINIT

	.PUSHJ	@[ICONS]	; initialize console

	NL0001			; successful init, return code 1

; return to our regularly scheduled initialization
XINIT:	CXF	1
	.POPJ

	.PAGE

;
; IFPGA: initialize FPGA
;
IFPGA:	CLA			; set up error return
	DCA	FPGAERR		; (assume no error)

	MM3			; access IOB address space

	CDF_STATUS		; check the FPGA INIT signal
	TAD	@0
	AND	[STAT_INIT]
	SNA CLA			; if it's low,
	 JMP	 FPGAE4		; ... FPGA not ready for bitstream

; copy fields 8..31 of the ROM to the FPGA

	TAD	[-24.]		; 24 fields
	DCA	X0
	TAD	[8.]		; first DAR field
	DCA	X1
	DCA	X2		; starting at offset 0

PROGPG: CDF_STATUS		; check for early finish
	TAD	@0
	AND	[STAT_DONE]
	SZA			; if it's high,
	 JMP	 FPGADC		; ... FPGA thinks it's configured

	TAD	X1		; set ROM field
	LDAR

PROGRM:	CDF_ROM			; fetch byte from ROM
	TAD	@X2
	CDF_FPGA		; send to FPGA
	DCA	@0

	ISZ	X2		; next in field
	 JMP	 PROGRM

	ISZ	X1		; next field
	ISZ	X0		; last field?
	 JMP	 PROGPG		; no, do next field

	CDF_STATUS		; check the FPGA DONE signal
	TAD	@0
	AND	[STAT_DONE]
	SNA			; if it's low,
	 JMP	 FPGAE5		; ... bitstream did not work

; now we do a self-test using IOTs that the FPGA should now be
; implementing

; the first test uses the TESTSIG IOT to make sure that SKIP, C0, and C1
; are properly connected and behave as expected

FPGADC:	NL0004			; ping C0 only
	TESTSIG
	 JMP	 .+2		; should be no skip
	JMP	 FPGAE6
	SZA			; AC should have been cleared
	 JMP	 FPGAE7

	NL0002			; ping C1 only
	TESTSIG
	 JMP	 .+2		; should be no skip
	JMP 	FPGAE6
	TAD	[-5252]		; 5252 should have been read
	SZA
	 JMP	 FPGAE8

	NL0001			; ping SKIP only
	TESTSIG
	 JMP	 FPGAE6		; should skip
	TAD	[-0001]		; AC should be untouched
	SZA
	 JMP	 FPGAE8

; the second test uses the TESTBUS IOT to make sure that all 12 data bus
; lines are connected and behave as expected

	TAD	[4000]
BUSTST:	DCA	X1
	TAD	X1
	TESTBUS			; should return complement
	TAD	X1		; now should be 7777
	IAC
	SZA
	 JMP	 FPGAE9
	TAD	X1     		; check next bit
	RTR
	SNA			; until whole word tested
	 JMP	 BUSTST

	CLA			; everything appears OK!
	JMP	RFPGA

FPGAE9:	ISZ	FPGAERR		; bus connection failure
FPGAE8:	ISZ	FPGAERR		; C1 failure
FPGAE7:	ISZ	FPGAERR		; C0 failure
FPGAE6:	ISZ	FPGAERR		; skip failure
FPGAE5:	ISZ	FPGAERR		; DONE inactive
FPGAE4:	ISZ	FPGAERR         ; INIT signal active
	NL0003
	TAD	FPGAERR
RFPGA:	MM2
	.POPJ

; Initialize console.  If the IOT1 GAL supports changing the SBC serial address,
; and the jumper is present, swap the the VT52 and the serial port.
; We can tell if the GAL has been updated because the PRISLU IOT will also
; set AC=0.

ICONS:  STA 			; make sure AC is non-zero
	PRISLU			; reset to primary address
	SZA CLA
	 .POPJ			; return if the AC did not clear

	TAD	[KEY]		; set the VT52 to the secondary address
	FUNLOCK                 ; in case jumper was removed w/o power off
	TAD	[3637]		; => ports 36/37
	FSETCON

	MM3			; get the status port
	CDF_STATUS
	CLA
	TAD	@0
	MM2

	AND	[STAT_JUMPER]	; get the jumper
	SZA			; and if it's not installed (1)
	 .POPJ			 ; ... leave the console at the serial port

; jumper is installed, switch console to VT52

	SECSLU			; set SBC6120 serial console to 36/37
	TAD	[KEY]
	FUNLOCK
	TAD	[0304]		; set VT52 to ports 03/04
	FSETCON

	.POPJ			; done!

	.PAGE

;
; INITVECT: copy vectors that the extension ROM needs from
; the tables found in BTS6120
;
INITVECT: TAD	[PATCHTAB-1]
	DCA	AX1
	CDF_HERE
INVE1:	TAD	@AX1
	SNA
	 .POPJ
	DCA	AX2
	TAD	@AX1
	JMS	GETV
	DCA	@AX2
	JMP	INVE1

;
; INI9HOOK: BTS6120 will jump here once it has done its basic
; initialization and printed the banner and device information.
; It will do this because the first phase of IOB initialization
; patched in a jump to here.
;
INI9HOOK:
;	.PUSHJ	@[HOOKCMDS]	; add in new commands

	CDF_HERE

	CLA
	TAD	[CFCMSG-1]	; say
	.PUSHJ	@[TASCIZ]	; "CFC: "

	TAD	CFSIZE		; check size
	SZA
	 JMP	 SHOWCF
	TAD	[CNPMSG-1]	; say "No CF Card"
	JMP	CFEMG		; but since we can hot plug we
				; will use a more neutral term

SHOWCF:	TAD	[-1]
	JMS	PJ0F2
PTDEC:	0
	TAD	[CSZMSG-1]	; say
	.PUSHJ	@[TASCIZ]	; "MB - "

	DCA	@[SECBUF+94.]	; make ID string ASCIZ
	TAD	[SECBUF+54.-1]	; and print it

CFEMG:	.PUSHJ	@[TASCIZ]
	.PUSHJ	@[TCRLF]

	CLA			; is the FPGA alive?
	TAD	FPGAERR
	SZA
	 JMP	INI9B		; no, only perform inits that don't use it

	TAD	[SAMSG-1]	; say
	.PUSHJ	@[TASCIZ]	; "IOB: ROM V"
	TAD	[VERSION]
	JMS	PJ0F2		; then the version
PTOCT4:	0
	TAD	[S2MSG-1]	; say
	.PUSHJ	@[TASCIZ]	; ", FPGA V"
	FVERSION		; IOT to get FPGA version
	JMS	PJ0F2		; and print that
PTOCT4B: 0
	.PUSHJ	@[TCRLF]

	;*** add other FPGA-related inits here ***
	.PUSHJ	@[HOOKCMDS]	; add in new commands
	
INI9B:	ISZ	PRESTA		; since it's a hook, the stored value is x-1
	CXF	0               ; but we just want to use it as a jmp vector
	JMP	@PRESTA		; and start the monitor

;
; TCRLF: print a newline
;
TCRLF:	JMS	PJ0F2
PCRLF:	0			; patched with BTS6120 CRLF routine
	.POPJ

;
; TASCIZ: print an ASCII string terminated by 0
;
TASCIZ:	DCA	ASCP		; save the pointer to the string
TASCI1:	TAD	@ASCP		; and get the first character
	SNA			; is this the end of the string ??
	 .POPJ 			; yes -- quit now
	JMS	PJ0F2		; no -- type this character
POUTCHR: 0			; patched with BTS6120 print char routine
	JMP	TASCI1		; and then loop until the end

	.PAGE
;
; CFHOOKS: intercept DISKRD and DISKWR so that we can seamlessly
; access the CF card via a partition number mapping
;
CFHOOKS:
	CDF_BTS1

	TAD	PDISKRD
	DCA	AX0
	TAD	[DISKRDHOOK]
	.PUSHJ	INTCPT
	DCA	PDISKRD

	TAD	PDISKWR
	DCA	AX0
	TAD	[DISKWRHOOK]
	.PUSHJ	INTCPT
	DCA	PDISKWR

; Also disable NODISK so that we can do our own checking
; otherwise the BTS6120 disk commands would balk if there
; was no IDE drive installed. It would be nice if we could
; check here, but NODISK is called before the partition
; number has been parsed.

	CDF_BTS0

	TAD	PNODISK
	DCA	AX0
	TAD	[CLA CLL]
	DCA	@AX0
	TAD	[6225] ; [.POPJ]
	DCA	@AX0

	CDF_HERE
	.POPJ

;
; INTCPT: patch one routine that has 3 words at AX0+1 with the address in AC
;
INTCPT:	.PUSH
	CLA
	TAD	[CIF_HERE]	; set up CIF 2 / JMP @[.+1] / <AC>
	DCA	@AX0

	NL0002			; compute address of constant
	TAD	AX0
	AND	[0177]		; make into a JMP @.+1
	TAD	[5600]
	DCA	@AX0

	.POP			; and store the address
	DCA	@AX0

	TAD	AX0		; and return next code address
	IAC
	.POPJ

	.PAGE

;
; INITCFC: check for a drive.  This is only used at startup to print
; a nice message, really, since it's also allowed to hot-plug a CF
; card.
;

INITCFC:
	CLA			; start out assuming that it's not present

;	MM3
;	CDF_CF
;	TAD	[340]		; set drive
;	DCA	@[CF_LB3]
;	TAD	[020]		; recalibrate
;	DCA	@[CF_CMD]

	DCA	CFSIZE
	DCA	CFPART		; zero partition and block
	DCA	CFRBN

	JMS	@[CFRW]		; call the generic CF I/O routine
	CMDIDD			; with the IDE identify command
	RDTRANS			; and the Read loop

	SZL			; successful?
	 .POPJ			;  no - could not identify

	NLM2			; aka [-2]
	TAD	@[SECBUF+10.]	; make sure sector size is 512
	SZA
	 .POPJ

;   Drives report the total number of LBA addressable sectors in words
; 60 and 61.  Sectors are 512 bytes, so simply dividing this value by 2048
; gives us the total drive size in Mb.  This code patches together twelve
; bits out of the middle of this doubleword, after throwing away the least
; significant 11 bits to divide by 2048.  This allows us to determine the
; size of drives up to 4Gb in a single 12 bit word.
	TAD	@[SECBUF+120.]	; get the high byte of the low word
	RAR			; throw away the 3 least significant
	RTR			; ...
	AND	[37]		; keep just 5 bits from this byte
	DCA	CFSIZE		; save it for a minute
	TAD	@[SECBUF+123.]	; get the low byte of the high word
	RTL			; left justify the seven MSBs of it
	RTL			; ...
	RAL			; ...
	AND	[7740]		; ...
	TAD	CFSIZE		; put together all twelve bits
	DCA	CFSIZE

	.POPJ

	.TITLE Disk Read and Write
;
; DISKRDHOOK: all calls to DISKRD in BTS6120 come here instead.  Here we
; either return to the original IDE code or use the new CF code, depending
; on the partition number.  If a call is made on a non-existent IDE drive,
; we return an error code rather than bombing out to the monitor as used to
; happen.  That was based on the idea that if you had booted, you knew
; that a drive must exist.  Now, there is hot plugging and there is also
; the possibility of booting from one drive and trying to access the other,
; so an error return is preferable.
; n.b. note that the method of storing 12-bit words in 2 8-bit bytes is
; defined by BTS6120 so that we can interoperate with the IDE drive,
; for example copying partitions to and fro. 
;
DISKRDHOOK:
	DCA	BUFSIZ		; save size
	CLA CLL
	TAD	@PDKPART	; get partition and see if it's
	TAD	[-CFPARITION]	; in the CF zone
	SNL
	 JMP	 IDERD		; nope, must be for SBC IDE

	DCA	CFPART		; save partition number on CF

	TAD	@PDKRBN
	DCA	CFRBN

	JMS	@[CFRW]		; call generic CF I/O routine
	CMDRDS			; with IDE read command
	RDTRANS			; and read loop

	SZL			; error?
	 JMP	 DRDXIT		 ; yes - could not read

; we have a sector in SECBUF, now assemble words from it into the destination

	CDF_BTS1

	TAD	[SECBUF-1]	; from the sector buffer
	DCA	AX0

	TAD	@PBUFPTR	; to the passed-in destination
	DCA	AX1

	TAD	@PBUFCDF	; in the passed-in field
	DCA	CDFRD

	TAD	@PBUFPNL	; and the passed in main/panel space
	IAC
	SNA
	 TAD	 [SPD-CPD+1]	; make CPD or SPD from it
	TAD	[CPD-1]
	DCA	RDPD

	CDF_HERE
RDXFR:	TAD	@AX0		; make words from the bytes in the buffer
	BSW
	CLL RTL
	TAD	@AX0
CDFRD:	0
RDPD:	0
	DCA	@AX1		; and store them in the user's buffer
	SPD
	CDF_HERE
	ISZ	BUFSIZ
	 JMP	 RDXFR

	CDF_BTS1
	TAD	AX1		; update the output pointer
	DCA	@PBUFPTR	; for multi-block reads

	CLA CLL			; and return OK
DRDXIT:	CXF	1
	.POPJ

IDERD:	CLA			; make sure an IDE drive exists
	TAD	@PDKSIZE
	SNA CLA
	 JMP	 @[NODISK]	; no, size was 0

	TAD	BUFSIZ		; chain to the original BTS6120
	CXF	1		; DISKRD
	JMP	@PDISKRD

; RDTRANS: low-level read loop from CF to SECBUF
RDTRANS: 0
RDLP:	MM3
	CDF_CF
	TAD	@[CF_DATA]	; flip bytes as they come in
	MQL			; due to the way the CF 8-bit
	TAD	@[CF_DATA]	; interface works
	MM2
	CDF_HERE
	DCA	@RAMPTR
	ISZ	RAMPTR
	MQA
	DCA	@RAMPTR
	ISZ	RAMPTR
	 JMP	 RDLP
	JMP	@RDTRANS

	.PAGE

;
; DISKRDHOOK: all calls to DISKWE in BTS6120 come here instead.  All
; the discussion about DISKRD applies here too.
;
DISKWRHOOK:
	DCA	BUFSIZ		; save size
	CLA CLL
	TAD	@PDKPART	; get partition and see if it's
	TAD	[-CFPARITION]	; in the CF zone
	SNL
	 JMP	 IDEWR		; nope, must be for SBC IDE

	DCA	CFPART		; save partition number on CF

	TAD	@PDKRBN
	DCA	CFRBN

; we want to assemble a sector in SECBUF from words in the source buffer

	CDF_BTS1

	TAD	@PBUFPTR	; from the user buffer
	DCA	AX0

	TAD	@PBUFCDF	; in the user field
	DCA	WRXFR

	TAD	@PBUFPNL	; in the user memory area (panel/main)
	IAC
	SNA
	 TAD	 [SPD-CPD+1]
	TAD	[CPD-1]
	DCA	WRPD

	TAD	[SECBUF-1]	; to the sector buffer
	DCA	AX1

WRXFR:	0
WRPD:	0
	TAD	@AX0		; get the buffer word
	MQL
	SPD
	CDF_HERE
	MQA
	BSW
	RTR
	AND	[17]
	DCA	@AX1		; and break it into bytes
	MQA
	DCA	@AX1
	ISZ	BUFSIZ
	 JMP	 WRXFR

	CDF_BTS1
	TAD	AX0		; update the destination pointer
	DCA	@PBUFPTR	; for multi-block writes

	JMS	@[CFRW]		; call the generic CF I/O routine
	CMDWRS			; with the write sector command
	WRTRANS			; and the write loop

DWRXIT:	CXF	1		; and return status returned from CFRW
	.POPJ

IDEWR:	CLA			; make sure an IDE drive exists
	TAD	@PDKSIZE
	SZA CLA
	 JMP	 HAVWDSK	 ; yes, size was non-zero

NODISK:	STA STL			; no, return -1 error
	CXF	1
	.POPJ

HAVWDSK:TAD	BUFSIZ		; chain to the original BTS6120
	CXF	1		; DISKWR
	JMP	@PDISKWR

; WRTRANS: write loop from SECBUF to the CF card
WRTRANS: 0
WRLP:	TAD	@RAMPTR		; reverse byte order during
	MQL			; transfer to match 8-bit access
	ISZ	RAMPTR		; method
	TAD	@RAMPTR
	MM3
	CDF_CF
	DCA	@[CF_DATA]
	MQA
	DCA	@[CF_DATA]
	MM2
	CDF_HERE
	ISZ	RAMPTR
	 JMP	WRLP
	JMP	@WRTRANS

	.PAGE

; CFRW: generic CF driver code.
;   entry: JMS, CFPART, CFRBN
;	inline parameters:
;	  Command	CF (IDE) command
;	  Handler	routine (JMS) to call for sector transfer
;			  called with SPD / MM2 / CDF_HERE
;			  should return with same
;   exit: L=AC=0 on success, L=1 AC=code on error

CFRW:	0			; call w/ JMS, follow by cmd, transfer vector

	NL0002			; get return address
	TAD	CFRW
	DCA	RWRET

; make sure a card is present

	MM3			; access status port
	CDF_STATUS

	TAD	@[0]
	AND	[STAT_CFDETECT]	; check the CF insert line
	SNA
	 JMP	HAVECF
	NL7777			; Error 7777 - no CF card
	JMP	RWERR1

; wait for CF ready

HAVECF:	CDF_CF
	TAD	[-100.]
	DCA	CFTO+1
	DCA	CFTO		; init timeout counter
WTRDY1:	TAD	@[CF_STAT]
	AND	[STSRDY+STSBSY]	; test the RDY & BSY flags
	TAD	[-STSRDY]	; wait for RDY set and BSY clear
	SNA CLA			; well?
	 JMP	 RDY1		 ; yes, ready!

	ISZ	CFTO
	 JMP	 WTRDY1
	ISZ	CFTO+1
	 JMP	 WTRDY1

	NLM2
	JMP	RWERR1
RDY1:
; set LBA
	TAD	CFRBN		; get the lower 12 bits of of the LBA
	DCA	@[CF_LB0]	; set the low LBA byte to b0..b7
	TAD	CFRBN		; get it again
	BSW			; shift right eight bits
	RTR			; so that b8..b11 are in b0..b3
	AND	[17]		; isolate them
	MQL			; save temporarily
	TAD	CFPART		; get the disk partition number
	AND	[17]		; isoloate the 4 lsb's
	CLL RTL			; shift them up into the 4 msb's
	RTL			; ...
	MQA			; and combine with the 4 msb's of the block
	DCA	@[CF_LB1]	; set the middle LBA byte to that

	TAD	CFPART		; get the disk partition number again
	RTR			; get the 8 msb's into b0..b7
	RTR			; ...
	DCA	@[CF_LB2]	; and store as the high LBA byte

	TAD	[340]		; select the master drive, 512 byte sectors,
	DCA	@[CF_LB3]	; LBA mode, and store in highest LBA byte
	; there are more block number bits in there, but we can't use them

	NL0001			; read or write 1 sector only
	DCA	@[CF_CNT]

; issue command

	MM2
	CDF_HERE
	TAD	@CFRW		; get the command
	MM3
	CDF_CF
	DCA	@[CF_CMD]	; issue it

; wait for DRQ

WTDRQ1:	TAD	@[CF_STAT]
	CLL RAR			; test the error bit (AC11)
	SZL			;
	 JMP	 RWERR
	RAL			; no - restore the original status value
	AND	[STSDRQ]	; and test the DRQ flag
	TAD	[-STSDRQ]	; wait for DRQ set
	SZA CLA			; well?
	 JMP	 WTDRQ1		 ; no, no data is waiting or waited for

; call transfer loop

HAVEDRQ: TAD	[SECBUF]
	DCA	RAMPTR

	MM2
	CDF_HERE
	ISZ	CFRW		; point to transfer function vector
	TAD	@CFRW
	DCA	CFRW
	JMS	@CFRW		; and call it

; wait for ready again (important for write, but doesn't hurt read)

	MM3
	CDF_CF
WTRDY2:	TAD	@[CF_STAT]
	CLL RAR			; test the error bit (AC11)
	SZL			;
	 JMP	 RWERR
	RAL			; no - restore the original status value
	AND	[STSBSY+STSRDY]	; and test the BUSY and RDY flags
	TAD	[-STSRDY]	; wait for BUSY clear and RDY set
	SZA CLA			; well?
	 JMP	 WTRDY2		; no, not yet

	CLA CLL			; everything ok
RWRET0:	MM2
	CDF_HERE
	JMP	@RWRET

RWERR:	CLA
	TAD	@[CF_ERR]
RWERR1:	STL			; there was an error - AC is code
	JMP	RWRET0

	.TITLE Tables, Messages
	.PAGE

; PATCHTAB: table of vectors that we want to fetch at startup and patch
; in our code.  Format is (destination-1), vector_number.
; While we are free here to mix field 0 and 1 vector numbers, the actual
; call to the function must use PJ0F2 or PJ1F2 as appropriate.
PATCHTAB:
	; general patches
	.DATA	PRESTA-1, VRESTA
	.DATA	POUTCHR-1, VOUTCHR
	.DATA	PCRLF-1, VCRLF
	.DATA	PTDEC-1, VTDECNW
	.DATA	AX0-1, VSYSIN9
	.DATA	PRESTA-1, VRESTA
	.DATA	PTOCT4-1, VTOCT4
	.DATA	PTOCT4B-1, VTOCT4
	.DATA	BTSHELP-1, VCMDTBL

	; CF patches
	.DATA	PDKSIZE-1, VDKSIZE
	.DATA	PPARMAP-1, VPARMAP
	.DATA	PDKPART-1, VDKPART
	.DATA	PDKRBN-1, VDKRBN
	.DATA	PBUFPTR-1, VBUFPTR
	.DATA	PBUFCDF-1, VBUFCDF
	.DATA	PBUFPNL-1, VBUFPNL
	.DATA	PDISKWR-1, VDISKWR
	.DATA	PDISKRD-1, VDISKRD
	.DATA	PNODISK-1, VNODISK

	; disassembler patches
	.DATA	PUIR-1, VUIR
	.DATA	PUPC-1, VUPC
	.DATA	PUFLAGS-1, VUFLAGS
	.DATA	PCR2-1, VCRLF
	.DATA	POUTC2-1, VOUTCHR
	.DATA	POCT3-1, VTOCT4
	.DATA	POCT4-1, VTOCT4
	.DATA	POUTC3-1, VOUTCHR
	.DATA	POUTC4-1, VOUTCHR
	.DATA	POCT5-1, VTOCT4
	.DATA	POCT6-1, VTOCT4
	.DATA	PTYPEIR-1, VTYPEIR
	.DATA	PRANGE-1, VRANGE
	.DATA	PADDR-1, VADDR
	.DATA	PADDRFLD-1, VADDRFLD
	.DATA	PHIGH-1, VHIGH
	.DATA	PHIGHFLD-1, VHIGHFLD

	.DATA	0

; Strings
SAMSG:	.ASCIZ	"IOB: ROM V"
S2MSG:	.ASCIZ	", FPGA V"
CFCMSG:	.ASCIZ	"CFC: "
CNPMSG:	.ASCIZ	"No CF Card"
CSZMSG:	.ASCIZ	"MB - "

	.PAGE

; HOOKCMDS: extend the BTS6120 commands.
; This function does 4 things:
;  1. Hooks the H command vector so that Help can be extended
;  2. Adds the X command at the end of the command list
;  3. Adds the XP command
;  4. Hooks the TYPEIR routine in order to display disassembly during TR

HOOKCMDS:
	CLA
	TAD	BTSHELP		; INITVECT put CMDTBL here
	DCA	X1		; now pointing at H(elp) entry

	CDF_BTS1
	ISZ	BTSHELP		; 1. move ptr to vector part
	TAD	@BTSHELP	; get current vector
	DCA	BTSHELP		; and save it

	TAD	[IOBHELP]	; change to our new Help implementation
	.PUSHJ	@[PATCHCMD]

EMPCMD: ISZ	X1		; search table for empty slots at end
	ISZ	X1
	TAD	@X1		; get cmd text
	SZA CLA			; 0 indicates end
	 JMP	 EMPCMD

	TAD	CMDX		; 2. Add X command
	DCA	@X1
	TAD	[CMDX+1]
	.PUSHJ	@[PATCHCMD]

	TAD	CMDXP		; 3. Add XP command
	DCA	@X1
	TAD	[CMDXP+1]
	.PUSHJ	@[PATCHCMD]

	TAD	PTYPEIR		; 4. intercept TYPEIR function
	DCA	AX0
	TAD	[TYPEIR]
	CDF_BTS0
	JMP	@[INTCPT]

; CMDX / CMDXP: the disassembler commands.  They share an implementation,
; since the only difference between them is whether main or panel memory is
; read.

CMDX:	.SIXBIT	/X /
	CLA
	TAD	[CPD]		; access main memory
	JMP	XCOMM

CMDXP:	.SIXBIT	/XP/
	CLA
	TAD	[SPD]		; access panel memory
XCOMM:	DCA	DPNL

	JMS	PJ0F2
PRANGE:	.BLOCK	1		; get range from input buffer

DA1L:	CDF_BTS0
	CLA CLL			; for L/LP commands, assume DF=IF
	TAD	@PADDRFLD
	RTR
	RAR
	.PUSH			; save the field for later printing
	TAD	@PADDRFLD
	DCA	DFIF		; the two lower digits of DFIF are the same

	TAD	@PADDR		; get the address from RANGE
	DCA	PC

	TAD	@PADDRFLD	; get the field from RANGE and make a CDF
	TAD	[CDF 0]
	DCA	DAF		; self-modify for memory space and field
	TAD	DPNL
	DCA	DAP
DAF:	0
DAP:	0
	TAD	@PC		; get the code word at the current address
	DCA	IR
	CDF_BTS0
	SPD

	.POP			; get back the field number
	.PUSHJ  DISA2		; and disassemble the instruction

	CDF_BTS0
	CLA	 		; are we at the end of the range?
	TAD	@PADDRFLD	; get the field
	CIA			; negate the field
	TAD	@PHIGHFLD	; compare to the high field
	SZA CLA	 		; are they equal ??
	 JMP INCR		; no -- continue
	TAD	@PHIGH		; yes -- compare the addresses
	CIA	 		;
	TAD	@PADDR		; ...
	SNA CLA			; are they equal ??
	 JMP XCMDDN		; yes -- stop

INCR:	ISZ	@PADDR		; increment address/field
	 JMP	 DA1L
	TAD	@PADDRFLD
	TAD	[10]
	DCA	@PADDRFLD
	JMP     DA1L

XCMDDN:	CXF	0		; done,
	.POPJ			; return to monitor

; DISA2: disassemble a line with all the trimmin's.  Displays code address,
; instruction value, and decoded instruction.  Enter with AC=field number.

DISA2:	AND	[7]		; turn field into ASCII
	TAD	["0"]
	JMS	PJ0F2		; and print it
POUTC4:	0
	TAD	PC		; print the PC
	JMS	PJ0F2
POCT5:	0
	TAD	[DS4MSG-1]	; say "/ "
	.PUSHJ	@[TASCIZ]
	TAD	IR		; print the IR (instruction)
	JMS	PJ0F2
POCT6:	0
	TAD	[DS5MSG-1]	; say "\t"
	.PUSHJ	@[TASCIZ]
	CDF_HERE
	JMP	@[DISA1]	; and print the disassembled line

	.TITLE Disassembler
	.PAGE

SCAN=AX0	; auto-incr ptr to OPJMP table
ASCAN=AX1	; auto-incr ptr to mnemonic table
MNEMP=X0	; address of last mnemonic output
MEA=X1		; effective address
MIA=X2		; indirect address

DISA1:	CLA
	DCA	MNEMP		; no mnemonics output yet

	TAD	IR		; get instruction
	BSW
	RTR			; extract upper digit
	AND	[16]
	TAD	[OPJMP-1]	; and make into table ptr
	DCA	SCAN

	TAD	@SCAN		; get pointer to mnemonic list for this group
	DCA	ASCAN		; SCAN now points to post-mnemonic routine (-1)

	TAD	IR		; do 1st mnemonic without sentinel check
	AND	@ASCAN		; so that opcode 0000 works
	JMP	MAT0B

MATCH0:	TAD	IR 		; get instruction
	AND	@ASCAN		; AND with mask
	SNA                     ; 0 mask at end of list
	 JMP	 MATCH8		 ; no more possible matches
MAT0B:	CIA
	TAD	@ASCAN		; compare masked IR with value
	SNA
	 JMP	 MATCH2		 ; matched!

; didn't match, skip over variable-length text to next entry
MATCH1:	TAD	@ASCAN		; because all strings are odd length
	AND	[77]		; just look at the even character
	SNA CLA			; for termination symbol (00)
	 JMP	 MATCH0		 ; yes, end of text, continue searching
	JMP	MATCH1		; no, skip another word

; did match, output the mnemonic and go back for more (for case of Operate)
MATCH2:	TAD	ASCAN		; save the location for later use in OPRGRP
	DCA	MNEMP
MATCH3:	TAD	@ASCAN		; get the packed 6-bit
	.PUSH			; save a couple of copies for later use
	.PUSH
	BSW			; print the first char
	.PUSHJ	TSIXB
	.POP			; print the second char
	.PUSHJ	TSIXB
	.POP
	AND	[77]		; was it a space? (n.b. space will have printed)
	SZA CLA
	 JMP	 MATCH3		 ; no, print next two
	JMP	MATCH0		; look for another mnemonic on this line

; done with mnemonic part of line
MATCH8:	TAD	@SCAN		; call post-mnemonic routine for this group
	DCA	MEA		; (Bad Things seem to happen when you .PUSHJ
	.PUSHJ	@MEA		;  indirect through an auto-increment location)
	CLA
	JMS	PJ0F2		; print newline
PCR2:	0
	.POPJ

TSIXB:	AND	[77]		; print a single 6bit character
	TAD	[" "]
	JMS	PJ0F2
POUTC2:	0
	.POPJ

; EAIGRP / EADGRP: post-mnemonic routines for memory reference instructions.
; prints the effective address (EA) and indirect data values.

EAIGRP:	TAD	DFIF		; for JMS/JMP, the EA uses IF
	RTR
	RAR
	SKP
EADGRP:	TAD	DFIF		; for AND/TAD/ISZ/DCA, the EA uses DF
EA2:	AND	[7]
	CLL R3L
	TAD	[CDF 0]		; make into a CDF
	DCA	@[EA3]		; and modify the fetch code

	TAD	DPNL		; get the CPD/SPD and modify the fetch code
	DCA	@[EA4]

	TAD	IR		; get in-page offset from the instruction
	AND	[177]
	MQL			; save it

	TAD	IR		; get ZP bit from instruction into Link
	R3L
	RTL

	CLA
	TAD	PC		; get current instruction page
	AND	[7600]

	SNL			; test ZP bit in instruction
	 CLA			 ; if ZP=0, clear page, otherwise keep PC page
	MQA			; combine with offset from instruction

	DCA	MEA		; now have effective addr

	.PUSHJ	@[EA3]		; get value at that address
	DCA	MIA

	TAD	IR		; get IA bit
	R3L
	RAL
	SNL CLA			; check if IA=0 (not indirect)
	 JMP	 EA6		 ; yes, do direct directly

	TAD	[DS2MSG-1]	; Say "@"
	.PUSHJ	@[TASCIZ]
	.PUSHJ	EA6		; print " [value]"
	TAD	MIA		; get double indirect value
	DCA	MEA
	.PUSHJ	@[EA3]
	DCA	MIA
	JMP	EA7		; print " [value]" and return

EA6:	TAD	MEA		; print the memory address
	JMS	PJ0F2
POCT3:	0
EA7:	TAD	[DS1MSG-1]	; print the indirect (or double-indirect)
	.PUSHJ	@[TASCIZ]	; value, in format "\t[xxxx]"
	TAD	MIA
	JMS	PJ0F2
POCT4:	0
	TAD	[DS6MSG-1]
	JMP	@[TASCIZ]

	.PAGE

; OPRGRP: called to finish up a mnemonic Operate instruction.
; It checks for instructions that have a parameter and also for unknown
; opcodes.

OPRGRP:	CLA CLL
	TAD	MNEMP		; get last mnemonic output
	SNA			; was there one?
	 JMP	 NOMNEM		 ; no - print "?" and return
	TAD	[-OP6NOCF]	; was it a CDF/CIF/CXF/PR?
	SZL CLA
	 .POPJ                   ; no - return w/o printing anything
	TAD	IR		; get instruction
	RTR			; and extract and print field number
	RAR
	AND	[7]
	TAD	["0"]
	JMS	PJ0F2
POUTC3:	0
	.POPJ

; OPRGRP: called to finish up a mnemonic IOT instruction
; Since IOTs are hardware defined, we have only a small standard set
; of them in the tables; so we see if anything useful was printed and,
; if not, print a '?'

IOTGRP:	TAD	MNEMP		; any mnemonics output?
	SZA CLA
	 .POPJ			 ; yes - return w/o printing anything
NOMNEM:	TAD	[DS3MSG-1]	; Say "?"
	JMP	@[TASCIZ]

; EA3: fetch word @MEA in correct memory space and field
; part of EAxGRP, and modifiied by that code

EA3:	CDF	0		; get value at that address
EA4:	CPD
	TAD	@MEA
	CDF_HERE
	SPD
	.POPJ

; PATCHCMD: generates a stub function in the BTS6120 Field 0 patch area.
; Used to add commands to CMDTBL.
; Inputs:
;	AC 	Field 2 function to be jumped to
;	X1      pointer in current field. [pointer+1] will be set to the stub
;		address.
; Outputs:
;	X1	incremented by 2
; Side effects:
;	F0PATCH	3 words used
;	DF=1
;	AC=0

PATCHCMD:
	.PUSH			; save the function address
	ISZ	X1		; move pointer to cmd vector part
	NL0002			; generate address of stub in patch area
	TAD	F0PATCH		; note that commands are in Field 0 even though
	DCA	@X1		; the table is in Field 1.
	ISZ	X1		; move pointer to next cmd string part
	.POP

	CDF_BTS0
	DCA	@F0PATCH	; store function address in patch area
	TAD	[CXF_HERE]	; stub starts with CXF 2
	DCA	@F0PATCH
	NLM1            	; then a jump through the function address,
	TAD	F0PATCH		; which is 2 back (but now 1 because of preincr)
	AND	[177]		; make into a JMP @
	TAD	[5600]
	DCA	@F0PATCH	; and add to stub
	CDF_BTS1
	.POPJ

; IOBHELP: Replacement Help command.  First calls the original, then
; prints our new help texts.

IOBHELP:
	TAD	BTSHELP		; get original help routine
	DCA	BTSHLP2
	JMS	PJ0F2		; and call it
BTSHLP2: 0

	TAD	[HLPLST-1]	; point to the list of help messages
	DCA	AX2		; in an auto index register
HLPLP:	TAD	@AX2		; get the next help message
	SNA			; end of list?
	 JMP	 HLPDN		; yes - we can quit now
	.PUSHJ	@[TASCIZ]
	.PUSHJ	@[TCRLF]	; and finish the line
	JMP	HLPLP		; keep typing until we run out of strings
HLPDN:	CXF	0		; and return to command interpreter
	.POPJ

; TYPEIR: Replacement TYPEIR function.  Displays a complete
; disassembly instead of IR>XXXX.

TYPEIR:	TAD	@PUPC		; get the current user environment
	DCA	PC		; into our local copies
	TAD	@PUFLAGS
	DCA	DFIF
	TAD	@PUIR
	DCA	IR
	TAD	[CPD]		; a TR always happens in main memory (?)
	DCA	DPNL

	TAD	DFIF		; get the IF for printing
	RTR
	RAR
	.PUSHJ	@[DISA2]	; and print the line

	CXF	0		; return to TR
	.POPJ

	.PAGE
	.TITLE Disassembler tables

; Opcode group tables.  For each group (upper digit), this
; table has a mnemonic table pointer and a pointer to a function
; that prints the parameters (e.g. effective address for AND)

OPJMP:	.DATA	OP0-1, EADGRP, OP1-1, EADGRP
	.DATA	OP2-1, EADGRP, OP3-1, EADGRP
	.DATA	OP4-1, EAIGRP, OP5-1, EAIGRP
	.DATA	OP6-1, OPRGRP, OP7-1, IOTGRP

; Mnemonic table. Each entry consists of a mask, a match value,
; and a text string in sixbit.  In order to detect the end of the
; string, it must be an odd number of characters long so that the
; last word is xx00.  .TEXT could have been used, but
; it always emits a full word of 0000 and so would have taken more space.

; the first 6 groups will always match, so no sentinels are necessary
; The Effective Address will be displayed for these
OP0:	.DATA	7000, 0000
	.SIXBIT	/AND/
OP1:	.DATA	7000, 1000
	.SIXBIT	/TAD/
OP2:	.DATA	7000, 2000
	.SIXBIT	/ISZ/
OP3:	.DATA	7000, 3000
	.SIXBIT	/DCA/
OP4:	.DATA	7000, 4000
	.SIXBIT	/JMS/
OP5:	.DATA	7000, 5000
	.SIXBIT	/JMP/

OP6:    	; mnemonics from here to OP6NOCF will have their 3rd digit
	.DATA	7707, 6201	; printed as the parameter
	.SIXBIT	/CDF/
	.DATA	7707, 6202
	.SIXBIT	/CIF/
	.DATA	7707, 6203
	.SIXBIT	/CXF/
	.DATA	7747, 6206
	.SIXBIT	/PR /
OP6NOCF:
	.DATA	7777, 6214
	.SIXBIT	/RDF/
	.DATA	7777, 6224
	.SIXBIT	/RIF/
	.DATA	7777, 6234
	.SIXBIT	/RIB/
	.DATA	7777, 6244
	.SIXBIT	/RMF/
	.DATA	7777, 6246
	.SIXBIT	/WSR/
	.DATA	7777, 6256
	.SIXBIT	/GCF/
	.DATA	7777, 6266
	.SIXBIT	/CPD/
	.DATA	7777, 6276
	.SIXBIT	/SPD/
	.DATA	7777, 6205
	.SIXBIT	/PPC1 /
	.DATA	7777, 6245
	.SIXBIT	/PPC2 /
	.DATA	7777, 6215
	.SIXBIT	/PAC1 /
	.DATA	7777, 6255
	.SIXBIT	/PAC2 /
	.DATA	7777, 6225
	.SIXBIT	/RTN1 /
	.DATA	7777, 6265
	.SIXBIT	/RTN2 /
	.DATA	7777, 6235
	.SIXBIT	/POP1 /
	.DATA	7777, 6275
	.SIXBIT	/POP2 /
	.DATA	7777, 6207
	.SIXBIT	/RSP1 /
	.DATA	7777, 6227
	.SIXBIT	/RSP2 /
	.DATA	7777, 6217
	.SIXBIT	/LSP1 /
	.DATA	7777, 6237
	.SIXBIT	/LSP2 /
	.DATA	7777, 6207
	.SIXBIT	/RSP1 /

	.DATA	7777, 6000
	.SIXBIT	"SKON/PRS "
	.DATA	7777, 6001
	.SIXBIT	/ION/
	.DATA	7777, 6002
	.SIXBIT	/IOF/
	.DATA	7777, 6003
	.SIXBIT	"SRQ/PGO"
	.DATA	7777, 6004
	.SIXBIT	"GTF/PEX"
	.DATA	7777, 6005
	.SIXBIT	/RTF/
	.DATA	7777, 6006
	.SIXBIT	/SGT/
	.DATA	7777, 6007
	.SIXBIT	/CAF/

	; console IOTs
	.DATA	7777, 6030
	.SIXBIT	/KCF/
	.DATA	7777, 6031
	.SIXBIT	/KSF/
	.DATA	7777, 6032
	.SIXBIT	/KCC/
	.DATA	7777, 6034
	.SIXBIT	/KRS/
	.DATA	7777, 6035
	.SIXBIT	/KIE/
	.DATA	7777, 6036
	.SIXBIT	/KRB/
	.DATA	7777, 6040
	.SIXBIT	/SPF/
	.DATA	7777, 6041
	.SIXBIT	/TSF/
	.DATA	7777, 6042
	.SIXBIT	/TCF/
	.DATA	7777, 6044
	.SIXBIT	/TPC/
	.DATA	7777, 6045
	.SIXBIT	/TSK/
	.DATA	7777, 6046
	.SIXBIT	/TLS/
	0

OP7:
	; Group 1
GRP1:	.DATA	7600, 7200
	.SIXBIT	/CLA/
	.DATA	7500, 7100
	.SIXBIT	/CLL/
	.DATA	7440, 7040
	.SIXBIT	/CMA/
	.DATA	7420, 7020
	.SIXBIT	/CML/
	.DATA	7401, 7001
	.SIXBIT	/IAC/
	.DATA	7416, 7002
	.SIXBIT	/BSW/
	.DATA	7416, 7004
	.SIXBIT	/RAL/
	.DATA	7416, 7006
	.SIXBIT	/RTL/
	.DATA	7416, 7010
	.SIXBIT	/RAR/
	.DATA	7416, 7012
	.SIXBIT	/RTR/
	.DATA	7416, 7014
	.SIXBIT	/R3L/
	.DATA	7777, 7000
	.SIXBIT	/NOP/
		; Group 1 composite
	.DATA	7777, 7240
	.SIXBIT	/(STA)/
	.DATA	7777, 7041
	.SIXBIT	/(CIA)/
	.DATA	7777, 7120
	.SIXBIT	/(STL)/
	.DATA	7777, 7204
	.SIXBIT	/(GLK)/
	.DATA	7777, 7300
	.SIXBIT	/(NL0000) /
	.DATA	7777, 7301
	.SIXBIT	/(NL0001) /
	.DATA	7777, 7326
	.SIXBIT	/(NL0002) /
	.DATA	7777, 7325
	.SIXBIT	/(NL0003) /
	.DATA	7777, 7307
	.SIXBIT	/(NL0004) /
	.DATA	7777, 7327
	.SIXBIT	/(NL0006) /
	.DATA	7777, 7340
	.SIXBIT	/(NLM1) /
	.DATA	7777, 7344
	.SIXBIT	/(NLM2) /
	.DATA	7777, 7346
	.SIXBIT	/(NLM3) /
GRP2:
	.DATA	7511, 7500
	.SIXBIT	/SMA/
	.DATA	7451, 7440
	.SIXBIT	/SZA/
	.DATA	7431, 7420
	.SIXBIT	/SNL/
	.DATA	7511, 7510
	.SIXBIT	/SPA/
	.DATA	7451, 7450
	.SIXBIT	/SNA/
	.DATA	7431, 7430
	.SIXBIT	/SZL/
	.DATA	7601, 7600
	.SIXBIT	/CLA/
	.DATA	7405, 7404
	.SIXBIT	/OSR/
	.DATA	7403, 7402
	.SIXBIT	/HLT/
	.DATA	7777, 7400
	.SIXBIT	/NOP/
	.DATA	7777, 7410
	.SIXBIT	/SKP/
GRP3:
	.DATA	7601, 7601
	.SIXBIT	/CLA/
	.DATA	7501, 7501
	.SIXBIT	/MQA/
	.DATA	7421, 7421
	.SIXBIT	/MQL/
	.DATA	7777, 7521
	.SIXBIT	/SWP/
	.DATA	7777, 7621
	.SIXBIT	/CAM/
	.DATA	7777, 7701
	.SIXBIT	/ACL/

	0

; Disassembler messages

DS1MSG:	.ASCIZ	"\t["
DS2MSG:	.ASCIZ	"@"
DS3MSG:	.ASCIZ	"?"
DS4MSG:	.ASCIZ	"/ "
DS5MSG:	.ASCIZ	"\t"
DS6MSG:	.ASCIZ	"]"

HLPLST:	.DATA	HELP0-1, HELP1-1, HELP2-1, HELP3-1, 0
HELP0:	.ASCIZ	//
HELP1:	.ASCIZ	/EXTENSION ROM COMMANDS/
HELP2:	.ASCIZ	/X  aaaaa[-bbbbb]\t\t-> Disassemble main memory/
HELP3:	.ASCIZ	/XP aaaaa[-bbbbb]\t\t-> Disassemble panel memory/

	.TITLE Sector Buffer
	; Sector buffer - 512 words
	.PAGE   34
	; Don't move this... the last word has to be at the end of the field
	; so that ISZ detects its end

SECBUF: .BLOCK  128.
	.PAGE
	.BLOCK  128.
	.PAGE
	.BLOCK  128.
	.PAGE
	.BLOCK  128.

	.END

